package org.jutils.init;

/*------------------------------------------------------------------------------
Name:      Property.java
Project:   jutils.org
Copyright: jutils.org, see jutils-LICENSE file
Comment:   Handling property variables for your project
Version:   $Id: Property.java,v 1.30 2000/11/08 11:04:29 winkler Exp $
------------------------------------------------------------------------------*/
import org.jutils.JUtilsException;
import org.jutils.text.StringHelper;
import org.jutils.io.FileUtil;

import java.io.*;
import java.util.Properties;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.util.Vector;
import java.net.URL;
import java.applet.Applet;


/**
 * Handling properties for your project, e.g. your property file $HOME/cool.properties.
 * <p />
 * This class searches the given property file in typical directories (see findPath() method).
 * The properties can be overwritten by properties given as a String[], for example
 * from the command line (see addArgs2Props() method for an explanation).
 * <p />
 * All environment variable (e.g. ${java.io.tmpdir}="/tmp", ${file.separator}="/") can be used
 * in the properties file and are replaced on occurrence.<br />
 * Further, you can use any of your own variables from the properties file or from the
 * command line with curly braces: ${...}.<br />
 * Example:<br />
 * A user may specify the MOMS_NAME and use it later as a variable as follows:<br />
 * <pre>   java -DMOMS_NAME=Joan ...</pre><br />
 * or in the properties file:<br />
 * <pre>   MOMS_NAME=Joan</pre>
 * or on the command line:<br />
 * <pre>   java org.MyApp  -MOMS_NAME Joan</pre><br />
 * and now use it like<br />
 * <pre>   MOMS_RECIPY=Strawberry cake from ${MONS_NAME} is the best.</pre>
 * which results to<br />
 * <pre>   MOMS_RECIPY=Strawberry cake from Joan is the best.</pre>
 * The maximum replacement (nesting) depth is 50.
 * <p />
 * A user may specify the PROJECT_HOME environment variable, the property file will be
 * searched there as well.
 * <p />
 * NOTE: These are often helpful variables
 * <ul>
 *    <li>${user.dir} =   The current directory </i>
 *    <li>${user.home} =  The users home directory</i>
 * </ul>
 * user.dir and user.home are without a path separator at the end
 * <p />
 * The property file settings are the weakest, they are overwritten
 * by system property settings and finally by command line arguments.
 *
 * <p />
 * The property file is searched for in the following order:
 * <ol>
 *    <li>Command line parameter '-propertyFile', e.g. "-propertyFile /tmp/xy.properties"</li>
 *    <li>Local directory user.dir</li>
 *    <li>In $PROJECT_HOME, e.g. "/opt/myproject"</li>
 *    <li>In user.home - $HOME</li>
 *    <li>In CLASSPATH - $CLASSPATH from your environment</li>
 *    <li>In java.home/lib/ext directory, e.g. "/opt/jdk1.2.2/jre/lib/ext"<br />
 *        You may use this path for Servlets, demons etc.</li>
 *    <li>In java.home directory, e.g. "/opt/jdk1.2.2/jre/lib"<br />
 *        You may use this path for Servlets, demons etc.</li>
 *    <li>Fallback: "\jutils" on DOS-WIN$ or "/usr/local/jutils" on UNIX</li>
 * <ol>
 * <p />
 * JDK 1.1 or higher only.
 *
 * @author Marcel Ruff
 */
public class Property
{   private final static String ME = "Property";

   private static String separator = null;
   /** The users home directory */
   private static String userHome = null;
   /** current directory, can it change for each instance or is it final in the JVM? */
   private String currentPath = null;
   /** user supplied path in PROJECT_HOME */
   private String projectHomePath = null;
   /** The java home ext directory, e.g. /opt/jdk1.2.2/jre/lib/ext */
   private static String javaHomeExt = null;
   /** The java home directory, e.g. /opt/jdk1.2.2/jre/lib */
   private static String javaHome = null;
   /** The supplied file name, e.g. cool.properties */
   private String propertyFileName = null;
   /** Holding all found properties */
   private Properties jutilsProperties = null;
   private final Properties dummyProperties = new Properties();
   /** Scan system environment variables? */
   private boolean scanSystemProperties = true;
   /** Replace occurrences of ${key} with the value of key */
   private boolean replaceVariables = true;
   /** Was the option '--help' or '-help' or '-h' or '-?' specified? */
   private boolean wantsHelp = false;
   /** Applet handle or null */
   private Applet applet = null;



    /**
    * Construct a property container from supplied property file and args array.
    * <p />
    * @param fileName The property file name, e.g. "project.properties"
    * @param scanSystemProperties Scan System.getProperties() as well, you can
    *        add variable to JVM like: java -DmyName=Joe ...
    * @param args A String array with properties, usually from command line, e.g.
    *        java myApp -trace true -name Joe
    * @param replaceVariables true: replace occurrences of ${key} with the value of key
    */
    public Property(String fileName, boolean scanSystemProperties, String[] args, boolean replaceVariables) throws JUtilsException {
          init(fileName, scanSystemProperties, args, replaceVariables);
    }   



    /**
    * Construct a property container from supplied property file and enumeration.
    * <p />
    * Usually used by servlets (ServletConfig conf):<br />
    * <pre>
    *    Properties extraProps = new Properties();
    *    Enumeration e = conf.getInitParameterNames();
    *    while (e.hasMoreElements()) {
    *       String name = (String)e.nextElement();
    *       extraProps.put(name, conf.getInitParameter(name));
    *    }
    * </pre>
    *
    * @param fileName The property file name, e.g. "project.properties"
    * @param scanSystemProperties Scan System.getProperties() as well, you can
    *        add variable to JVM like: java -DmyName=Joe ...
    * @param args A String array with properties, usually from command line, e.g.
    *        java myApp -trace true -name Joe
    * @param replaceVariables true: replace occurrences of ${key} with the value of key
    */
    public Property(String fileName, boolean scanSystemProperties, java.util.Properties extraProps, boolean replaceVariables) throws JUtilsException {
           String[] arr = new String[2*extraProps.size()];
           Enumeration e = extraProps.propertyNames();
           int count = 0;
           while (e.hasMoreElements()) {
              String name = (String)e.nextElement();
              arr[count++] = "-" + name;
              arr[count++] = (String)extraProps.getProperty(name);
           }
          init(fileName, scanSystemProperties, arr, replaceVariables);
    }   



    /**
    * Construct a property container from supplied property file (URL) and Applet.
    * <p />
    * NOTE: Applet HTML PARAM values may not contain variables ${...}, they are currently not replaced.
    *
    * @param propertyFileUrl applet.getCodeBase()+"xy.properties" e.g. "http://myHost.com:80/xy.properties"
    *
    * @param scanSystemProperties Scan System.getProperties() as well, you can
    *        add variable to JVM like: java -DmyName=Joe ...<br />
    *   NOTE: If you set this to true, you need a signed applet!
    *
    * @param applet Not yet supported. How to access all applet parameters???
    * @param replaceVariables true: replace occurrences of ${key} with the value of key
    */
    public Property(String propertyFileUrl, boolean scanSystemProperties, Applet applet, boolean replaceVariables) throws JUtilsException {
          System.out.println("Property for Applet: propertyFileUrl=" + propertyFileUrl + ", scanSystemProperties=" + scanSystemProperties + ", replaceVariables=" + replaceVariables);
          this.scanSystemProperties = scanSystemProperties;
          this.replaceVariables = replaceVariables;
          this.applet = applet;
    
          // Variant downloading it with a jar file using HTML ARCHIVE tag?
          //  java.net.URL      url;
          //  url = this.getClass().getResource(propertyFileName);
    
          if (propertyFileUrl == null) {
             loadProps(null, null);
             if (jutilsProperties == null)
                jutilsProperties = dummyProperties;
             return;
          }
    
          InputStream inputStream = null;
          URL url = null;
          try {
             url = new URL(propertyFileUrl);
             this.propertyFileName = url.getFile();
             try {
                inputStream = url.openStream();
             }
             catch (java.io.IOException e) {
                System.out.println("Property file " + url + " read error, continuing without it: " + e.toString());
                //throw new JUtilsException(ME + ".Error", "URL access: " + e.toString());
             }
          }
          catch (java.net.MalformedURLException e) {
             System.out.println("Property file " + propertyFileUrl + " is malformed, continuing without it: " + e.toString());
          }
          loadProps(inputStream, null);
          if (jutilsProperties == null)
             jutilsProperties = dummyProperties;
    }   



    /**
        * Internal Initialize.
        */
    private void init(String fileName_, boolean scanSystemProperties, String[] args, boolean replaceVariables) throws JUtilsException {
          System.out.println("Property: filenName=" + fileName_ + ", scanSystemProperties=" + scanSystemProperties + ", args=" + (args==null ? "null" : "["+args.length+"]") + ", replaceVariables=" + replaceVariables);
          // System.out.println(org.jutils.log.Log.getStackTrace());
          this.propertyFileName = fileName_;
          this.scanSystemProperties = scanSystemProperties;
          this.replaceVariables = replaceVariables;
          this.separator = System.getProperty("file.separator");
          this.userHome = System.getProperty("user.home");
          this.currentPath = System.getProperty("user.dir");
          this.projectHomePath = System.getProperty("PROJECT_HOME");
          this.javaHomeExt = System.getProperty("java.ext.dirs");
          this.javaHome = System.getProperty("java.home");
    
          FileInputStream fis = null;
          try {  // Read user supplied properties file
             if (propertyFileName != null) {
                String fileName = findFile(propertyFileName, args);
                if (fileName != null) {
                   File file = new File(fileName);
                   fis = new FileInputStream(file);
                }
                else {
                   System.err.println("Property: Please copy " + propertyFileName + " to your home directory. We continue with default settings.");
                }
             }
             loadProps(fis, args);
             if (fis != null) fis.close();
          }
          catch (IOException e) {
             throw new JUtilsException(ME + ".Error", "Unable to initialize " + propertyFileName + ": " + e);
          }
          if (jutilsProperties == null)
             jutilsProperties = dummyProperties;
    }   



    /**
    * Get the internal handle.
    * @return The Properties handle.
    */
    public final java.util.Properties getProperties() {
          return jutilsProperties;
    }   



    private String get_(String key)
    {
          if (applet != null) {
             String a = applet.getParameter(key);
             if (a != null) return a;
          }
          return (String)jutilsProperties.get(key);
    }   



    /**
        * Set or overwrite a property,
        * note that currently no variable replacement is implemented for the passed value.
        * @param key The key for this property
        * @param value The value for it
        * @return The value, ${...} variables are replaced
        */
    public final String set(String key, String value) throws JUtilsException
    {
          value = replaceVariable(key, value);
          jutilsProperties.setProperty(key, value);
          return value;
    }   



    /**
    * Try to find the given key.
    * @param key the key to look for
    * @param defaultVal the default value to return if key is not found
    * @return The String value for the given key
    */
    public final String get(String key, String defaultVal) {
          String str = get_(key);
          if (str == null)
             return defaultVal;
          return str;
    }   



    /**
    * Try to find the given key.
    * <p />
    * Example:<br />
    * NameList=Josua,David,Ken,Abel<br />
    * Will return each name separately in the array.
    * @param key the key to look for
    * @param defaultVal The default value to return if key is not found
    * @param separator  The separator, typically ","
    * @return The String array for the given key, the elements are trimmed (no leading/following white spaces)
    */
    public final String[] get(String key, String[] defaultVal, String separator) {
          String str = get_(key);
          if (str == null)
             return defaultVal;
          StringTokenizer st = new StringTokenizer(str, separator);
          int num = st.countTokens();
          String[] arr = new String[num];
          int ii=0;
          while (st.hasMoreTokens()) {
             arr[ii++] = st.nextToken().trim();
          }
          return arr;
    }   



    /**
    * Try to find the given key.
    * @param key the key to look for
    * @param defaultVal the default value to return if key is not found
    * @return The int value for the given key
    */
    public final int get(String key, int defaultVal) {
          String str = get_(key);
          if (str == null)
             return defaultVal;
          try {
             return Integer.parseInt(str);
          }
          catch (Exception e) {
             return defaultVal;
          }
    }   



    /**
    * Try to find the given key.
    * @param key the key to look for
    * @param defaultVal the default value to return if key is not found
    * @return The long value for the given key
    */
    public final long get(String key, long defaultVal) {
          String str = get_(key);
          if (str == null)
             return defaultVal;
          try {
             return Long.parseLong(str);
          }
          catch (Exception e) {
             return defaultVal;
          }
    }   



    /**
    * Try to find the given key.
    * <p />
    * See toBool() for a list of recognized strings.
    * @param key the key to look for
    * @param defaultVal the default value to return if key is not found
    * @return The boolean value for the given key
    */
    public final boolean get(String key, boolean defaultVal) {
          String str = get_(key);
          if (str == null)
             return defaultVal;
          try {
             return toBool(str);
          }
          catch (Exception e) {
             return defaultVal;
          }
    }   



    /**
    * Try to find the given key.
    * <p />
    * @param key the parameter key to look for
    * @return true if the property exists
    */
    public final boolean propertyExists(String key) {
          String str = get_(key);
          if (str == null)
             return false;
          return true;
    }   



    /**
    * Remove the given property.
    * <p />
    * This method does nothing if the key is not in the property hashtable.
    * @param key the key to remove
    */
    public final Object removeProperty(String key) {
          return jutilsProperties.remove(key);
    }   



    /**
    * If set to true user wants you to display a usage text.
    * @return true if the option '--help' or '-help' or '-h' or '-?' was given.
    */
    public final boolean wantsHelp() {
          return wantsHelp;
    }   



    /**
    * Use this method only the first time to initialize everything.
    * @args This key/value parameter array is added to the poperties object (see addArgs2Props()).
    * @return the initialized Properties
    */
    public final Properties loadProps(InputStream inputStream, String[] args) throws JUtilsException {
          if (jutilsProperties != null) {
             System.out.println("Property: Reloading Property file ... args=" + args);
          }
    
          // set default
          jutilsProperties = new Properties();
          try {
             // 1. Read user supplied properties file
             if (inputStream != null) {
                jutilsProperties.load(inputStream);
                System.out.println("Property: Loaded file " + propertyFileName);
             }
             else
                System.out.println("Property: No property file given.");
    
             // 2. Read system environment, e.g. java -Dname=joe
             if (scanSystemProperties == true) {
                Properties env = System.getProperties();
                for (Enumeration e = env.propertyNames(); e.hasMoreElements();) {
                   String key = (String) e.nextElement();
                   jutilsProperties.put(key, System.getProperty(key));
                }
             }
             else
                System.out.println("Property: No system properties scanned.");
    
             // 3. read user supplied String[], e.g. command line parameters
             if (args != null) {
                addArgs2Props(jutilsProperties, args);
             }
             else
                System.out.println("Property: No args array given.");
    
    
             // 4. Substitute dynamic variables, e.g. ${user.dir}
             if (replaceVariables == true)
                replaceVariables();
          }
          catch (IOException e) {
             throw new JUtilsException(ME + ".Error", "Unable to initialize " + propertyFileName + ": " + e);
          }
          return jutilsProperties;
    }   



    /**
        * Replace all dynamic variables, e.g. ${XY} with their values.
        */
    private final void replaceVariables() throws JUtilsException {
          if (replaceVariables == false)
             return;
          for (Enumeration e = jutilsProperties.keys(); e.hasMoreElements();) {
             String key = (String) e.nextElement();
             String value = get_(key);
             jutilsProperties.put(key, replaceVariable(key, value));
          }
    }   



    /**
        * Replace dynamic variables, e.g. ${XY} with their values.
        * <p />
        * The maximum replacement (nesting) depth is 50.
        * @param key
        * @value The value string which may contain zero to many ${...} variables
        * @return The new value where all ${} are replaced.
        */
    private final String replaceVariable(String key, String value) throws JUtilsException {
          if (replaceVariables == false)
             return value;
          String origValue = value;
          for (int ii = 0;; ii++) {
             int from = value.indexOf("${");
             if (from != -1) {
                int to = value.indexOf("}", from);
                if (to == -1) {
                   throw new JUtilsException(ME + ".InvalidVariable", "Invalid variable '" + value.substring(from) + "', expecting ${} syntax.");
                }
                String sub = value.substring(from, to + 1); // "${XY}"
                String subKey = sub.substring(2, sub.length() - 1); // "XY"
                String subValue = get_(subKey);
                if (subValue == null) {
                   throw new JUtilsException(ME + ".UnknownVariable", "Unknown variable " + sub + "");
                }
                value = StringHelper.replaceAll(value, sub, subValue);
             }
             else {
                if (ii > 0) {
                   System.out.println("Property: Replacing '" + key + "=" + origValue + "' to '" + value + "'");
                }
                return value;
             }
             if (ii > 50) {
                throw new JUtilsException(ME + ".MaxNest", "Maximum nested depth of 10 reached for variable '" + get_(key) + "'.");
             }
          }
    }   



    /**
    * Parse a string to boolean.
    * <p />
    * @param token for example "false"
    * @return true for one of "true", "yes", "1", "ok"<br />
    *         false for "false", "0", "no"
    * @exception if none of the above strings
    */
    static public final boolean toBool(String token) throws JUtilsException {
          if (token == null)
             throw new JUtilsException(ME, "Can't parse <null> to true or false");
          if (token.equalsIgnoreCase("true") || token.equalsIgnoreCase("1") || token.equalsIgnoreCase("ok") || token.equalsIgnoreCase("yes"))
             return true;
          if (token.equalsIgnoreCase("false") || token.equalsIgnoreCase("0") || token.equalsIgnoreCase("no"))
             return false;
          throw new JUtilsException(ME, "Can't parse <" + token + "> to true or false");
    }   



    /**
    * Look for properties file.
    * <p />
    * The property file is searched in the specific order as described above
    * @param fileName e.g. "cool.properties"
    * @param args     Only e.g. "-propertyFile /tmp/xy.properties" is evaluated (if given)
    * @return The path to file, e.g. "\jutils\"
    */
    public final String findPath(String fileName, String[] args) {
          if (fileName == null)
             return null;
    
          File f = null;
    
          /* Coded in findFile()
          try {
             String argsLocation = Args.getArg(args, "-propertyFile", (String)null);
             if (argsLocation != null) {
                f = new File(argsLocation);
                if (f.exists()) {
                   return argsLocation;
                }
                else
                   System.out.println("Property: File '" + fileName + "' not given on command line '-propertyFile'");
             }
          } catch(JUtilsException e) {
          }
          */
    
          f = new File(currentPath, fileName);
          if (f.exists()) {
             System.out.println("Property: Loading " + fileName + " from directory " + currentPath);
             return currentPath;
          }
          else
             System.out.println("Property: File '" + fileName + "' is not in current directory " + currentPath);
    
    
          if (projectHomePath == null) {
             System.out.println("Property: File '" + fileName + "' is not in PROJECT_HOME, 'java -DPROJECT_HOME=...' is not set ...");
          }
          else {
             if (!projectHomePath.endsWith(separator))
                projectHomePath += separator;
             f = new File(projectHomePath, fileName);
             if (f.exists()) {
                System.out.println("Property: Loading " + fileName + " from directory " + projectHomePath);
                return projectHomePath;
             }
             else
                System.out.println("Property: File '" + fileName + "' is not in directory PROJECT_HOME=" + projectHomePath);
          }
    
          f = new File(userHome, fileName);
          if (f.exists()) {
             System.out.println("Property: Loading " + fileName + " from directory " + userHome);
             return userHome;
          }
          else
             System.out.println("Property: File '" + fileName + "' is not in user.home directory " + userHome);
    
          ClassLoader loader = Property.class.getClassLoader();
          if (loader != null) {
             java.net.URL url = loader.getResource(fileName);
             if (url != null ) {
                String fn = url.getFile(); // path and filename
                f = new File(fn);
                if (f.exists()) {
                   System.out.println("Property: Loading " + fileName + " from CLASSPATH directory " + f.getParent());
                   return f.getParent();
                }
             }
          }
          System.out.println("Property: File '" + fileName + "' is not in CLASSPATH");
    
          f = new File(javaHomeExt, fileName);
          if (f.exists()) {
             System.out.println("Property: Loading " + fileName + " from directory " + javaHomeExt);
             return javaHomeExt;
          }
          else
             System.out.println("Property: File '" + fileName + "' is not in java.home/lib/ext directory " + javaHomeExt);
    
    
          f = new File(javaHome, fileName);
          if (f.exists()) {
             System.out.println("Property: Loading " + fileName + " from directory " + javaHome);
             return javaHome;
          }
          else
             System.out.println("Property: File '" + fileName + "' is not in java.home directory " + javaHome);
          String guess;
          if (separator.equals("/"))
             guess = "/usr/local/jutils/";
          else
             guess = "\\jutils\\";
    
    
          f = new File(guess, fileName);
          if (f.exists()) {
             System.out.println("Property: Loading " + fileName + " from directory " + guess);
             return guess;
          }
          else
             System.out.println("Property: File '" + fileName + "' is not in directory " + guess);
    
          return null;
    }   



    /**
    * Find properties file.
    * <p />
    * See findPath() for search - logic
    *
    * @param fileName e.g. "cool.properties"
    * @param args Only "-propertyFile /tmp/xy.properties" is evaluated (if given)
    * @return The path to file, e.g. "\jutils\"
          */
    public final String findFile(String fileName, String[] args) {
    
          try {
             String argsLocation = Args.getArg(args, "-propertyFile", (String)null);
             if (argsLocation != null) {
                File f = new File(argsLocation);
                if (f.exists()) {
                   propertyFileName = argsLocation;
                   return argsLocation;
                }
                else
                   System.out.println("Property: File '" + fileName + "' not given on command line '-propertyFile'");
             }
          } catch(JUtilsException e) {
          }
    
          if (fileName == null) {
             System.out.println("Property: No property file specified.");
             return null;
          }
          String path = findPath(fileName, args);
          if (path == null) {
             System.err.println("Property: File '" + fileName + "' not found");
             return null;
          }
          String fullName = FileUtil.concatPath(path, fileName);
          // System.out.println("Property: Using file '" + fullName + "'");
          return fullName;
    }   



    /**
    * Add key/values, for example from startup command line args to
    * the property variable.
    * <p />
    * Args parameters are stronger and overwrite the property file variables.
    * <p />
    * The arg key must have a leading - or -- (as usual on command line).<br />
    * The leading - are stripped (to match the property variable)
    * args must be a tuple (key / value pairs).
    * <p />
    * Example:
    * <pre>
    *   java org.jutils.Main  isCool true -iorPort 3400
    * </pre>
    * This would overwrite a properties file variable "iorPort" (if there was one)
    * and set the variable isCool to "true"
    */
    public void addArgs2Props(String[] args) throws JUtilsException {
          addArgs2Props(jutilsProperties, args);
          replaceVariables();
    }   



    /**
    * See method addArgs2Props(args)
    */
    private void addArgs2Props(Properties props, String[] args) throws JUtilsException {
          if (args == null)
             return;
          for (int ii = 0; ii < args.length; ii++) {
             String arg = args[ii];
             if (arg.startsWith("-")) { // only parameters starting with "-" are recognized
                String key = arg;
                if (arg.equals("--help") || arg.equals("-?") || arg.equals("-h") || arg.equals("-help")) {
                   wantsHelp = true;
                   continue; // ignore
                }
                if (arg.startsWith("--"))
                   key = arg.substring(2); // strip "--"
                else
                   if (arg.startsWith("-"))
                      key = arg.substring(1); // strip "-"
                if ((ii + 1) >= args.length) {
                   // System.err.println("Property: The property '" + arg + "' has no value");
                   throw new JUtilsException(ME + ".MissingValue", "The property '" + arg + "' has no value.");
                }
                String arg2 = args[ii + 1];
                props.put(key, arg2);
                ii++;
                // System.out.println("Property: Setting command line argument " + key + " with value " + arg2);
             }
             else {
                throw new JUtilsException(ME + ".Invalid", "Ignoring unknown argument <" + arg + ">, it should start with '-'.");
             }
          }
    }   



    /**
    * Dump all properties to xml representation.
    */
    public String toXml() {
          StringBuffer buf = new StringBuffer();
          buf.append("<Property>\n");
          for (Enumeration e = jutilsProperties.propertyNames(); e.hasMoreElements();) {
             String key = (String) e.nextElement();
             buf.append("   <").append(key).append(">");
             buf.append(StringHelper.replaceAll((String) jutilsProperties.get(key), "<", "&lt;"));
             buf.append("</").append(key).append(">\n");
          }
          buf.append("</Property>\n");
          return buf.toString();
    }   



    /**
    * For testing only
    * <p />
    * <pre>
    * java org.jutils.init.Property -Persistence.Driver myDriver -isCool true -xml "<hello>world</hello>"
    * java org.jutils.init.Property -uu "You are the user \${user.name}" -double "\${uu} using \${java.vm.name}"
    * java org.jutils.init.Property -NameList Josua,David,Ken,Abel
    * java org.jutils.init.Property -urlVariant true -url http://localhost/xy.properties
    * java org.jutils.init.Property -hashVariant true
    * </pre>
    */
    public static void main(String args[]) {
          String ME = "Property";
    
          try {
             boolean urlVariant = Args.getArg(args, "-urlVariant", false);
             if (urlVariant) {
                {
                   System.out.println("*** Test 1");
                   Property props = new Property(null, true, (Applet)null, true); // initialize
                   System.out.println("TestVariable=" + props.get("TestVariable", "ERROR"));
                   System.out.println("java.home=" + props.get("java.home", "ERROR"));
                }
                {
                   System.out.println("*** Test 2");
                   String url = Args.getArg(args, "-url", "http://localhost/xy.properties");
                   Property props = new Property(url, true, (Applet)null, true); // initialize
                   System.out.println("TestVariable=" + props.get("TestVariable", "ERROR"));
                   System.out.println("java.home=" + props.get("java.home", "ERROR"));
                }
                System.exit(0);
             }
          } catch (JUtilsException e) {
             //System.err.println(e.toXml(true));
             System.err.println(e.toString());
             System.exit(0);
          }
    
          try {
             boolean hashVariant = Args.getArg(args, "-hashVariant", false);
             if (hashVariant) {
                {
                   System.out.println("*** Test 1");
                   Properties extra = new Properties();
                   extra.put("city", "Auckland");
                   extra.put("country", "NewZealand");
                   Property props = new Property(null, true, extra, true); // initialize
                   System.out.println("city=" + props.get("city", "ERROR"));
                   System.out.println("country=" + props.get("country", "ERROR"));
                   System.out.println("TestVariable=" + props.get("TestVariable", "shouldBeUndef"));
                   System.out.println("java.home=" + props.get("java.home", "ERROR"));
                }
                System.exit(0);
             }
          } catch (JUtilsException e) {
             //System.err.println(e.toXml(true));
             System.err.println(e.toString());
             System.exit(0);
          }
    
          try {
             System.out.println("yes=" + Property.toBool("yes"));
             System.out.println("no=" + Property.toBool("no"));
             System.out.println("aaa=" + Property.toBool("aaa"));
          }
          catch (Exception e) {
             System.out.println("OK, aaa=ERROR");
          }
          System.out.println("\n\n");
          try {
             Property props = new Property("xmlBlaster.properties", true, args, true); // initialize
    
             System.out.println("Persistence=" + props.get("Persistence", false));
             System.out.println("Persistence.Dummy=" + props.get("Persistence.Dummy", false));
             System.out.println("Persistence.Driver=" + props.get("Persistence.Driver", "NONE"));
             System.out.println("Persistence.Dummy=" + props.get("Persistence.Dummy", "NONE"));
             System.out.println("\n\n");
             Properties raw = props.getProperties();
             System.out.println("All properties:\n" + raw);
             System.out.println("All properties as XML:\n" + props.toXml());
             String[]arr = { "Mohamed", "Charly" }; // default
             arr = props.get("NameList", arr, ",");
             for (int ii=0; ii<arr.length; ii++)
                System.out.println("NameList[" + ii + "]=" + arr[ii]);
    
             System.out.println("\nChanging 'Persistence.Dummy' to  '[[${java.version}]]' dynamically:");
             props.set("Persistence.Dummy", "[[${java.version}]]");
             System.out.println("Persistence.Dummy=" + props.get("Persistence.Dummy", "NONE"));
          }
          catch (JUtilsException e) {
             //System.err.println(e.toXml(true));
             System.err.println(e.toString());
          }
    }   



    /**
     * The default constructor. Deeded for some IDEs to work properly.
     * Creation date: (07.11.2000 20:14:53)
     */
    public Property() {
        System.out.println("Property : Default constructor called.");
    }}
