package org.jutils.text;

import java.util.*;
/**
 * A nice and fine "StringVector" that refits dynamically and can store Strings without casting!
 * Use is much like Vector, althougt it can not be casted to a vector. To transform, use one of that
 * toVector methods.
 * <p>------------------------------------------------------------------------------<br>
 * Name:     	StringList<br>
 * Project:  	jutils<br>
 * Copyright:	jutils.org, see license file<br>
 * Comment:		a nice and fine "StringVector" that refits dynamically and can store Strings without casting!
 * Creation date: (05.07.2000 14:50:41)   <br>
 * Version:   <pre>$Id: StringList.java,v 1.4 2000/09/01 17:55:35 winkler Exp $</pre> <br>
 * ------------------------------------------------------------------------------<p>
 * @author: M.Winkler, doubleSlash GmbH
 */
public class StringList implements java.io.Serializable {

   /** The default capacity of this list. */
   public static final int DEFAULT_INITIAL_SIZE = 10;

   /** Determines that a default comparison should be done: case sensitive, no trimming */
   public static final int DEFAULT_COMPARE = 0;

   /** When comparing a string entry, then this is done by considering cases. */
   public static final int IGNORE_CASE_ON_COMPARE = 1;

   /** When comparing a string entry of this list, trim leading and trailing spaces. */
   public static final int TRIM_ON_COMPARE = 2;

   /** When comparing a string entry of this list, only references are compared. all other aspects are ignored. */
   public static final int ONLY_COMPARE_REFERENCE_ON_COMPARE = Integer.MAX_VALUE;

   /** The default capacity raise when more capacity is used.*/
   public static final int DEFAULT_RAISE = 10;
   private String[] data_ = null;
   private int count_ = 0;
   private int raise_ = DEFAULT_RAISE;



    /**
    * StringList constructor comment.
    */
    public StringList() {
        super();
        data_ = new String[DEFAULT_INITIAL_SIZE];
    }



    /**
    * Creates a new StringList by copying the strings of the provided array.
    * Creation date: (05.07.2000 15:02:43)<br>
    * @author: M.Winkler, doubleSlash GmbH
    *
    * @param strings java.lang.String[]
    */
    public StringList(String[] strings) {
        if (strings == null)
            throw new IllegalArgumentException("null arrays cannot be processed by this constructor, sorry.");
        int initialSize = Math.max(DEFAULT_INITIAL_SIZE, strings.length);
    }



    /**
    * Creation date: (05.07.2000 14:52:22)<br>
    * @author: M.Winkler, doubleSlash GmbH
    *
    * @param initialSize int
    */
    public StringList(int initialSize) {
        if (initialSize < 1)
            throw new IllegalArgumentException("StringLists must have an initial capacity > 0!");
        data_ = new String[initialSize];
    }



    /**
    * Adds a new String to the list.
    * Creation date: (05.07.2000 15:26:44)<br>
    * @author: M.Winkler, doubleSlash GmbH
    *
    * @param newString java.lang.String
    */
    public void add(String newString) {

        // check parameters
        if (newString == null)
            throw new IllegalArgumentException("Cannot handle null strings!");

        // enough space available?!?
        if (count_ == data_.length)
            ensureCapacity(count_ + raise_);

        // put into
        synchronized (data_) {
            data_[count_] = newString;
            count_++;
        }
    }



    /**
    * Creation date: (05.07.2000 14:59:00)<br>
    * @author: M.Winkler, doubleSlash GmbH
    *
    * @return int
    */
    public int capacity() {
        return data_.length;
    }



    /**
     * @author Michael Winkler, doubleSlash GmbH
     *
     */
    public synchronized void clear() {
        if (count_ > 0) {
            for (int index = count_ - 1; index > -1; index--) {
                data_[index] = null;
            }
            count_ = 0;
        }
    }



    /**
    * Does the same like compare (string, DEFAULT_OPTIONS_ON_COMPARE)
    * Creation date: (05.07.2000 15:34:15)<br>
    * @author: M.Winkler, doubleSlash GmbH
    *
    * @return boolean if true, this string already is in the list in respec6t to the options.
    * @param string java.lang.String a string to compare with.
    */
    public boolean contains(String string) {
        return contains(string, DEFAULT_COMPARE);
    }



    /**
    * Checks wether this String already is in the List.
    * Creation date: (05.07.2000 15:34:15)<br>
    * @author: M.Winkler, doubleSlash GmbH
    *
    * @return boolean if true, this string already is in the list in respec6t to the options.
    * @param string java.lang.String a string to compare with.
    * @param options int may be any bit mask that is either ONLY_COMPARE_REFERENCE_ON_COMPARE or OR'ed using those other *On_COMPARE constants.
    */
    public boolean contains(String string, int options) {

        // check arguments
        if (string == null)
            throw new IllegalArgumentException("StringList cannot handle null references, so you donnot have to search for them!");

        // define one
        boolean returnValue = false;

        // synchronize
        synchronized (data_) {

            // initializing local variables
            boolean onlyCompareReferences = options == ONLY_COMPARE_REFERENCE_ON_COMPARE;
            boolean ignoreCase = (options & IGNORE_CASE_ON_COMPARE) > 0;
            boolean trim = (options & TRIM_ON_COMPARE) > 0;
            String stringOnIndex = null;
            if (!onlyCompareReferences) {
                if (trim)
                    string = string.trim();
                if (ignoreCase)
                    string = string.toUpperCase();
            }


            // do the job.
            int index = count_ - 1; // count backwards.
            while ((index > -1) && !returnValue) {
                stringOnIndex = data_[index];

                // only compare references?!?
                if (onlyCompareReferences) {
                    returnValue = stringOnIndex == string;
                }
                else { // compare content

                    // trimming?!?
                    if (trim)
                        stringOnIndex = stringOnIndex.trim();

                    // case sensitive?!?
                    if (ignoreCase)
                        stringOnIndex = stringOnIndex.toUpperCase();

                    // doing compare
                    returnValue = string.equals(stringOnIndex);
                } // end of else

                // decrement counter
                index--;
            } // end of while
        } // end of synchronized
        return returnValue;
    }



    /**
    * Creation date: (05.07.2000 14:58:38)<br>
    * @author: M.Winkler, doubleSlash GmbH
    *
    * @return int
    */
    public int count() {
        return count_;
    }



    /**
    * Drops the last element in the list. This operation is quite fast.
    * Creation date: (05.07.2000 18:28:58)<br>
    * @author: M.Winkler, doubleSlash GmbH
    *
    */
    public void dropLast() {
        synchronized (data_) {
            if (count_ > 0) {
                count_--;
                data_[count_] = null;
            }
        }
    }



    /**
    * Creation date: (05.07.2000 15:22:55)<br>
    * @author: M.Winkler, doubleSlash GmbH
    *
    * @param newMinCapacity int
    */
    public void ensureCapacity(int newMinCapacity) {
        synchronized (data_) {

            // check wethzer something should happen at all.
            if (newMinCapacity > data_.length) {
                String[] newData = new String[newMinCapacity];
                System.arraycopy(data_, 0, newData, 0, count_);
                data_ = newData;
            }
        }
    }



    /**
    * Returns the last added String in the list.
    * Creation date: (05.07.2000 18:31:26)<br>
    * @author: M.Winkler, doubleSlash GmbH
    *
    * @return java.lang.String
    */
    public String getLast() {
        String returnValue = null;
        synchronized (data_) {
            if (count_ > 0) {
                returnValue = data_[count_ - 1];
            }
        }
        return returnValue;
    }



    /**
    * Creation date: (05.07.2000 17:56:13)<br>
    * @author: M.Winkler, doubleSlash GmbH
    *
    * @return java.lang.String
    * @param index int
    */
    public String getStringAt(int index) {
        if ((index >= count_) || (index < 0))
            throw new IllegalArgumentException("Index out of bounds:" + index);
        return data_[index];
    }



    /**
    * Test this implementation.
    * Creation date: (05.07.2000 15:57:16)<br>
    * @author: M.Winkler, doubleSlash GmbH
    *
    * @param args java.lang.String[]
    */
    public static void main(String[] args) {
        final String SAMPLE_STRING = "aSampleString";
        final int STRINGS_COUNT = 500;


        // init locals
        int errorCount = 0;
        String[] instancesInList = new String[STRINGS_COUNT];
        String[] similarStrings = new String[STRINGS_COUNT];
        StringList list = new StringList();


        // generating random strings
        System.out.println("Generating a similar string pairs ...");
        for (int index = 0; index < STRINGS_COUNT; index++) {

            // create a string
            String aString = "" + (int) (STRINGS_COUNT * Math.random()) + SAMPLE_STRING;

            // put it into the array
            instancesInList[index] = aString;

            // put the same reference into the list
            list.add(aString);

            // do some small changes and put it into the array of the simmilar instances
            similarStrings[index] = " " + aString.toUpperCase() + " ";
        }

        // Sample output

        // test size
        // --------------------------------------------
        System.out.println("List size/capacity:" + list.count() + "/" + list.capacity());
        System.out.println("Checking size ...");
        if (list.count() != STRINGS_COUNT) {
            System.out.println("\tWrong number of elements in list!");
            errorCount++;
        }

        // test contains
        // --------------------------------------------
        System.out.println("Performing test of contains(String)...");

        // retrieve an instance
        int index = STRINGS_COUNT / 2;
        String aString = instancesInList[index];
        String aSimilarString = similarStrings[index];

        // reference check
        System.out.println("Testing contains() doing a reference check ...");

        // present instance (must always be found!)
        int options = StringList.ONLY_COMPARE_REFERENCE_ON_COMPARE;
        if (!list.contains(aString, options)) {
            System.out.println("\tReference not found!");
            errorCount++;
        }

        // similar but not identical reference  (may not be found)
        if (list.contains(aSimilarString, options)) {
            System.out.println("\tReference found that hasnt put into the list!");
            errorCount++;
        }

        // case insensitive check
        // --------------------------------------------
        options = StringList.IGNORE_CASE_ON_COMPARE;
        System.out.println("Testing contains() case insensitive ...");

        // present instance (must always be found!)
        if (!list.contains(aString.toUpperCase(), options)) {
            System.out.println("\tIdentical reference not found! ");
            errorCount++;
        }

        // similar but not identical reference (may not be found)
        if (list.contains(aSimilarString, options)) {
            System.out.println("\tSimilar string with other case found althougt case sensitivity was activated!");
            errorCount++;
        }

        // trim check
        // --------------------------------------------
        options = StringList.TRIM_ON_COMPARE;
        System.out.println("Testing contains() case sensitive but trimming put active ...");

        // present instance (must always be found!)
        if (!list.contains(aString.trim(), options)) {
            System.out.println("\tIdentical reference not found! ");
            errorCount++;
        }

        // similar but not identical reference (may not be found)
        if (list.contains(aSimilarString.trim(), options)) {
            System.out.println("\tSimilar string with other case found althougt case sensitivity was activated!");
            errorCount++;
        }

        // insensitive, trim check
        // --------------------------------------------
        options = StringList.TRIM_ON_COMPARE | StringList.IGNORE_CASE_ON_COMPARE;
        System.out.println("Testing contains() with case insensitive and trim ...");

        // present instance (must be found!)
        if (!list.contains(aString.trim(), options)) {
            System.out.println("\tIdentical reference not found! ");
            errorCount++;
        }

        // similar but not identical reference (may not be found)
        if (!list.contains(aSimilarString, options)) {
            System.out.println(" \tFully activated options must find the provided similar reference, but didn ' t !");
            errorCount++;
        }

        // test sorting
        // --------------------------------------------
        System.out.println("Sorting ...");
        list.sort(String.CASE_INSENSITIVE_ORDER);
        System.out.println("Checking sort ...");
        String oldString = list.getStringAt(0);
        String activeString = null;
        for (int lv = 1; lv < list.count(); lv++) {
            activeString = list.getStringAt(lv);
            int compareValue = String.CASE_INSENSITIVE_ORDER.compare(oldString, activeString);
            if (compareValue > 0) {
                System.out.println("Wrong sort order at " + (lv - 1) + "/" + lv + " : " + oldString + "<->" + activeString);
                errorCount++;
            }
            oldString = activeString;
        }



        // Doing sequence test
        // --------------------------------------------
        int referenceCheckOptions = StringList.ONLY_COMPARE_REFERENCE_ON_COMPARE;
        int fullToleratedCheckOptions = StringList.TRIM_ON_COMPARE | StringList.IGNORE_CASE_ON_COMPARE;
        System.out.println("Running sequencing test ...");
        for (index = 0; index < STRINGS_COUNT; index++) {
            if (!list.contains(instancesInList[index], referenceCheckOptions)) {
                System.out.println("\tMissed reference check!");
                errorCount++;
            }
            if (!list.contains(similarStrings[index], fullToleratedCheckOptions)) {
                System.out.println("\tMissed full tollerated check check!");
                errorCount++;
            }
        }

        // output
        System.out.println("------------------------------------------------------------------");
        System.out.println("Errors encountered : " + errorCount);

        // closing
        System.exit(0);
    }



    /**
    * Does a sort of the data stored. As a comparator,
    * <tt>String.CASE_INSENSITIVE_ORDER</tt> is used.
    * Creation date: (05.07.2000 17:32:48)<br>
    * @author: M.Winkler, doubleSlash GmbH
    *
    */
    public void sort() {
        synchronized (data_) {
            Arrays.sort(data_, 0, count_ - 1, String.CASE_INSENSITIVE_ORDER);
        }
    }



    /**
    * Does a sort of the data stored.
    * Creation date: (05.07.2000 17:32:48)<br>
    * @author: M.Winkler, doubleSlash GmbH
    *
    */
    public void sort(Comparator c) {
        synchronized (data_) {
            Arrays.sort(data_, 0, count_, c);
        }
    }



    /**
    * Copies all contained elements into the passed array.
    * Creation date: (05.07.2000 18:47:00)<br>
    * @author: M.Winkler, doubleSlash GmbH
    *
    * @return java.lang.String[]
    */
    public String[] toArray(String[] array) {
        synchronized (data_) {
            if (count_ > 0) {
                if (array == null)
                    array = new String[count_];
                System.arraycopy(data_, 0, array, 0, count_);
            }
        }
        return array;
    }
}

