package org.jutils.time;

/*------------------------------------------------------------------------------
Name:      Timeout.java
Project:   jutils.org
Copyright: jutils.org, see jutils-LICENSE file
Comment:   Allows you be called back after a given delay.
Version:   $Id: Timeout.java,v 1.4 2000/07/26 11:16:39 ruff Exp $
------------------------------------------------------------------------------*/
import org.jutils.JUtilsException;

import java.util.*;


/**
 * Allows you be called back after a given delay.
 * <p />
 * Note that this class should be called Timer, but with JDK 1.3 there
 * will be a java.util.Timer.
 * <p />
 * There is a single background thread that is used to execute the I_Timeout.timeout() callback.
 * Timer callbacks should complete quickly. If a timeout() takes excessive time to complete, it "hogs" the timer's
 * task execution thread. This can, in turn, delay the execution of subsequent tasks, which may "bunch up" and execute in
 * rapid succession when (and if) the offending task finally completes.
 * <p />
 * This singleton is thread-safe.
 * <p />
 * This class does not offer real-time guarantees, but usually notifies you within ~ 20 milliseconds
 * of the scheduled time.
 * <p />
 * Adding or removing a timer is good performing, also when huge amounts of timers (> 1000) are used.
 * <p />
 * TODO: Use a thread pool to dispatch the timeout callbacks.
 * <p />
 * Example:<br />
 * <pre>
 * public class MyClass implements I_Timeout {
 *   ...
 *   Timeout timeout = Timeout.getInstance();
 *   Long timeoutHandle = timeout.addTimeoutListener(this, 4000L, "myTimeout");
 *   ...
 *   public void timeout(Object userData) {
 *      // userData contains String "myTimeout"
 *      System.out.println("Timeout happened");
 *      ...
 *      // If you want to activate the timer again:
 *      timeoutHandle = timeout.addTimeoutListener(this, 4000L, "myTimeout");
 *   }
 *   ...
 *   // if you want to refresh the timer:
 *   timeoutHandle = timeout.refreshTimeoutListener(timeoutHandle, 1500L);
 *   ...
 * }
 * </pre>
 *
 * JDK 1.2 or higher only.
 *
 * @author ruff@swand.lake.de
 */
public class Timeout extends Thread
{
   /** Name for logging output */
   private static String ME = "Timeout";
   /** The singleton handle */
   private static Timeout theTimeout = null;
   /** Sorted map */
   private TreeMap map = null;
   /** Start/Stop the Timeout manager thread */
   private boolean running = true;
   /** To protect the singleton */
   private static final java.lang.Object SYNCHRONIZER = new java.lang.Object();

   /**
   * Helper holding the callback interface an some user data to be looped through.
   */
   private class Container
   {
     I_Timeout callback;
     Object userData;
     long creation;
     Container(I_Timeout callback, Object userData) {
       this.callback = callback;
       this.userData = userData;
       this.creation = System.currentTimeMillis();
     }
   }

   /**
   * Access to Timeout singleton
   */
   public static Timeout getInstance() {
      if (theTimeout == null) { // avoid 'expensive' synchronized
         synchronized (SYNCHRONIZER) {
            if (theTimeout == null) {
               theTimeout = new Timeout();
               theTimeout.setDaemon(true);
               theTimeout.start();
            }
         }
      }
      return theTimeout;
   }

   /**
   * private Constructor for Singleton Pattern
   */
   private Timeout() {
      super("Timeout-Thread");
      map = new TreeMap();
   }

   /**
   * Starts the Timeout manager thread.
   */
   public void run() {
      while (running) {
         long delay = 100000; // sleep veeery long
         synchronized (map) {
            try {
               Long nextWakeup = (Long) map.firstKey();
               long next = nextWakeup.longValue();
               long current = System.currentTimeMillis();
               delay = next - current;
               if (delay <= 0) {
                  Container container = (Container) map.remove(nextWakeup);
                  if (false) {
                     long time = System.currentTimeMillis();
                     long diff = time - nextWakeup.longValue();
                     System.out.println("Timeout occurred, calling listener with real time error of " + diff + " millis");
                  }
                  container.callback.timeout(container.userData);
                  continue;
               }
            }
            catch (NoSuchElementException e) {
               // The listener map is empty, nothing to do.
            }
         }
         try {
            synchronized (theTimeout) {
               theTimeout.wait(delay);
            }
         }
         catch (InterruptedException i) {
            // Wakeing up, and check if there is something to do
         }
      }
   }

   /**
   * Add a listener which gets informed after 'delay' milliseconds.
   * <p />
   * After the timeout happened, you are not registered any more. If you want to
   * cycle timeouts, you need to register again.
   *
   * @param listener Your callback handle (you need to implement this interface)
   * @param delay The timeout in milliseconds
   * @param userData Some arbitrary data you supply, it will be routed back to you when the timeout occurs through method I_Timeout.timeout()
   * @return A handle which you can use to unregister with removeTimeoutListener()
   */
   public final Long addTimeoutListener(I_Timeout listener, long delay, Object userData) {
      Long key = null;
      synchronized (map) {
         while (true) {
            key = new Long(System.currentTimeMillis() + delay);
            Object obj = map.get(key);
            if (obj == null) {
               map.put(key, new Container(listener, userData));
               break;
            }
            else {
               // We loop to avoid two similar keys, this should happen very seldom
               // As we don't guarantee real time, we don't care to be 1 milli inaccurate.
               // If we don't like this delay, we would need to use a multimap.
            }
         }
      }
      synchronized (theTimeout) {
         theTimeout.notify();
      }
      return key;
   }

   /**
   * Refresh a listener before the timeout happened.
   * <p />
   * NOTE: The returned timeout handle is different from the original one.
   *
   * @param key The timeout handle you received by a previous addTimeoutListener() call<br />
   *            It is invalid after this call.
   * @param delay The timeout in milliseconds measured from now.
   * @return A new handle which you can use to unregister with removeTimeoutListener()
   * @exception JUtilsException if key is invalid
   */
   public final Long refreshTimeoutListener(Long key, long delay) throws JUtilsException {
      Long newKey = null;
      synchronized (map) {
         Container container = (Container) map.remove(key);
         if (container == null) {
            throw new JUtilsException(ME, "The timeout handle '" + key + "' is unknown, no timeout refresh done");
         }
         while (true) {
            newKey = new Long(System.currentTimeMillis() + delay);
            Object obj = map.get(newKey);
            if (obj == null) {
               map.put(newKey, container);
               break;
            }
            // We loop to avoid two similar keys, this should happen very seldom
         }
      }
      synchronized (theTimeout) {
         theTimeout.notify();
      }
      return newKey;
   }

   /**
   * Remove a listener before the timeout happened.
   * <p />
   * @param key The timeout handle you received by a previous addTimeoutListener() call
   */
   public final void removeTimeoutListener(Long key) {
      synchronized (map) {
         Container container = (Container) map.remove(key);
         if (container != null) {
            container.callback = null;
            container.userData = null;
            container = null;
         }
      }
   }

   /**
   * Is this handle expired?
   * <p />
   * @param key The timeout handle you received by a previous addTimeoutListener() call<br />
   * @return true/false
   */
   public final boolean isExpired(Long key) {
      synchronized (map) {
         return map.get(key) == null;
      }
   }

   /**
   * How long to my timeout.
   *
   * @param key The timeout handle you received by a previous addTimeoutListener() call
   * @return Milliseconds to timeout, or -1 if not known
   */
   public final long spanToTimeout(Long key) {
      synchronized (map) {
         Container container = (Container) map.get(key);
         if (container == null) {
            return -1;
         }
         // We know that the key contains the timeout date
         return getTimeout(key) - System.currentTimeMillis();
      }
   }

   /**
   * How long am i running.
   *
   * @param key The timeout handle you received by a previous addTimeoutListener() call
   * @return Milliseconds since creation, or -1 if not known
   */
   public final long elapsed(Long key) {
      synchronized (map) {
         Container container = (Container) map.get(key);
         if (container == null) {
            return -1;
         }
         return System.currentTimeMillis() - container.creation;
      }
   }

   /**
   * Access the end of life span.
   * @param key The timeout handle you received by a previous addTimeoutListener() call
   * @return Time in milliseconds since midnight, January 1, 1970 UTC
   *         or -1 if not known
   */
   public final long getTimeout(Long key) {
      if (key == null)
         return -1;
      return key.longValue();
   }

   /**
   * Reset all pending timeouts.
   * <p />
   */
   public final void removeAll() {
      synchronized (map) {
         map.clear();
      }
   }

   /**
   * Reset this singleton, stop the Timeout manager thread.
   */
   public final void destroy() {
      removeAll();
      running = false;
      synchronized (theTimeout) {
         theTimeout.notify();
      }
      synchronized (SYNCHRONIZER) {
         theTimeout = null;
      }
   }

   /**
   * Only for testing.
   * <p />
   * Invoke: java -Djava.compiler= org.jutils.time.Timeout
   */
   public static void main(String args[]) throws Exception {
      String ME = "Timeout-Tester";
      Timeout timeout = Timeout.getInstance();

      //==== 1. We test the functionality:
      System.out.println("Phase 1: Testing basic functionality ...");

      // Test to remove invalid keys
      timeout.removeTimeoutListener(null);
      timeout.removeTimeoutListener(new Long(12));

      // We have the internal knowledge that the key is the scheduled timeout in millis since 1970
      // so we use it here for testing ...
      final Long[] keyArr = new Long[4];
      class Dummy1 implements I_Timeout {
         private String ME = "Dummy1";
         private int counter = 0;
         public void timeout(Object userData) {
            long time = System.currentTimeMillis();
            long diff = time - keyArr[counter].longValue();
            if (Math.abs(diff) < 40)

               // Allow 40 millis wrong notification (for code execution etc.) ...
               System.out.println("Timeout occurred for " + userData.toString() + " at " + time + " millis, real time failure=" + diff + " millis.");
            else
               System.err.println("*****ERROR: Wrong timeout occurred for " + userData.toString() + " at " + time + " millis, scheduled was " + keyArr[counter] + " , real time failure=" + diff + " millis.");
            counter++;
         }
      }
      Dummy1 dummy = new Dummy1();
      keyArr[2] = timeout.addTimeoutListener(dummy, 4000L, "timer-4000");
      keyArr[3] = timeout.addTimeoutListener(dummy, 5000L, "timer-5000");
      keyArr[3] = timeout.refreshTimeoutListener(keyArr[3], 5500L);
      keyArr[0] = timeout.addTimeoutListener(dummy, 1000L, "timer-1000");
      keyArr[1] = timeout.addTimeoutListener(dummy, 1000L, "timer-1000");
      long span = 0;
      if ((span = timeout.spanToTimeout(keyArr[2])) < 3000L)
         System.err.println("*****ERROR: This short span to timeout = " + span + " is probably wrong, or you have a very slow computer.");
      else
         System.out.println("Span to life of " + span + " is reasonable");
      Long key = timeout.addTimeoutListener(dummy, 1000L, "timer-1000");
      timeout.removeTimeoutListener(key);
      try {
         key = timeout.refreshTimeoutListener(key, 1500L);
      }
      catch (JUtilsException e) {
         System.out.println("Refresh failed which is OK (it is a test): " + e.reason);
      }
      if (timeout.isExpired(keyArr[2]))
         System.err.println("*****ERROR: Should not be expired");
      else
         System.out.println("Correct, is not expired");
      try {
         Thread.currentThread().sleep(7000L);
      }
      catch (Exception e) {
         System.err.println("*****ERROR: main interrupt: " + e.toString());
      }
      if (!timeout.isExpired(keyArr[2]))
         System.err.println("*****ERROR: Should be expired");
      else
         System.out.println("Correct, is expired");


      //===== 2. We test a big load:
      final int numTimers = 1000;
      System.out.println("Phase 2: Testing " + numTimers + " timeouts ...");
      timeout.destroy();
      timeout = Timeout.getInstance(); // get a new handle
      class Dummy2 implements I_Timeout {
         private String ME = "Dummy2";
         private int counter = 0;
         private long start = 0L;
         public void timeout(Object userData) {
            if (counter == 0) {
               start = System.currentTimeMillis();
            }
            counter++;
            if (counter == numTimers) {
               long diff = System.currentTimeMillis() - start;
               if (diff < 2000L)
                  System.out.println("Success, tested " + numTimers + " timers, all updates came in " + diff + " millis");
               else
                  System.err.println("*****ERROR: Error testing " + numTimers + " timers, all updates needed " + diff + " millis");
            }
         }
      }
      Dummy2 dummy2 = new Dummy2();
      long start = System.currentTimeMillis();
      for (int ii = 0; ii < numTimers; ii++) {
         timeout.addTimeoutListener(dummy2, 4000L, "timer-" + ii);
      }
      System.out.println("Feeding of " + numTimers + " done, " + (long) (1000 * (double) numTimers / (System.currentTimeMillis() - start)) + " adds/sec");
      System.out.println("Waiting in main");
      try {
         Thread.currentThread().sleep(10000L);
      }
      catch (Exception e) {
         System.err.println("*****ERROR:main interrupt: " + e.toString());
      }
      System.err.println("*****ERROR:Test failed");
   }
}
